
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

typedef unsigned char byte;
typedef unsigned short word;
typedef unsigned long ulong;

/********************************************************/
/****                                                ****/
/**** ALFL2PCX v1.0                                  ****/
/****                                                ****/
/**** Converts images from (Amiga, maybe more) LFL   ****/
/**** files to PCX                                   ****/
/****                                                ****/
/**** Code by Lloyd Rosen, except PCX routines.      ****/
/**** Contact: rosen@zenzer.net, www.zenzer.net      ****/
/****                                                ****/
/**** Also get LFL2PCX.                              ****/
/****                                                ****/
/********************************************************/

static void decode_pic(void);
static void pcx_writebitmap(byte *bmap, word width, word height);
static void pcx_writeline(byte *p, word len);
static void pcx_writepalette(void);
static void doseek(ulong offset);
static void doskip(ulong bytes);
static word getw(void);
static byte getb(void);
static void wrb(byte b);
static void wrwi(word w);

FILE *lfl_fh, *pcx_fh;

word gfxoff;


byte *lfl, *bitmap, decode_flag = 1;
word bmwidth, bmheight;
word roomwidth;

int main(int ac, char **av)
{
	int i;

	puts("AFL2PCX v1.0 - Convert (AMIGA) .LFL files to PCX images\nCode: Lloyd Rosen.");
	if (ac < 3) {
		puts("usage: alfl2pcx infile outfile [options]\n");
		puts("options: -d  (lfl file already decoded)");
		exit(0);
	}

	for (i = 3; i < ac; ++i) {
		if (av[i][0] == '-') {
			switch(av[i][1]) {
				case('d'):
					decode_flag = 0;
					break;
			}
			continue;
		}
	}

	if ((lfl_fh = fopen(av[1], "rb")) == NULL) {
		printf("Error opening LFL file %s\n", av[1]);
		exit(1);
	}

	if ((pcx_fh = fopen(av[2], "wb")) == NULL) {
		printf("Error opening PCX file %s\n", av[2]);
		fclose(lfl_fh);
		exit(1);
	}

	doseek(4);			/* skip to background info */

	roomwidth = getw() >> 3;
	bmheight = getw();

	doskip(2);
	gfxoff = getw();

	printf("Reading from %ld in %s... (width %ld, height %ld)\n", gfxoff, av[1], roomwidth, bmheight);

	doseek(gfxoff);

	bmwidth = roomwidth<<3;

/* allocate memory for the bitmap */

	bitmap = malloc(bmwidth*bmheight);
	memset(bitmap, 0, bmwidth*bmheight);

/* respect my authoritah! */

	decode_pic();

	if (lfl_fh) fclose(lfl_fh);

	pcx_writebitmap(bitmap, bmwidth, bmheight);
	if (pcx_fh) fclose(pcx_fh);
}

static byte b18184[128];

static void decode_pic(void)
{
	word z, row;
	byte col = 0, cnt = 1, dat;
	byte xd5;
	byte *rowptr;

	xd5 = 0;

	for (z = 0; z < bmwidth; ++z) {
		rowptr = b18184;

		for (row = 0; row < bmheight; ++row) {

			if (--cnt == 0) {
				dat = getb();
				if (dat & 0x80) {
					cnt = dat & 0x7f;
					xd5 = 1;
				} else {
					cnt = dat >> 4;
					xd5 = 0;
				}
				if (!cnt)
					cnt = getb();

				col = dat & 0xf;
			}

			if (!xd5)
				*rowptr = col;

			bitmap[row * bmwidth + z] = *rowptr++;
	  	}
	}
}



static word getw(void)
{
	byte b1 = fgetc(lfl_fh);
	byte b2 = fgetc(lfl_fh);

	if (decode_flag) {
		b1 ^= 0xff;
		b2 ^= 0xff;
	}
	return(word)(b1 | (b2<<8));
}

static byte getb(void)
{
	byte b = fgetc(lfl_fh);

	if (decode_flag)
		b ^= 0xff;
	return(b);
}

static void doseek(ulong offset)
{
	fseek(lfl_fh, offset, SEEK_SET);
}

static void doskip(ulong bytes)
{
	fseek(lfl_fh, bytes, SEEK_CUR);
}

static void wrb(byte b)
{
	fputc(b, pcx_fh);
}

static void wrwi(word w)
{
	fputc(w, pcx_fh);
	fputc(w >> 8, pcx_fh);
}



static void pcx_writebitmap(byte *bmap, word width, word height)
{
	word x;

	wrb(10);		// manufacturer always 10
	wrb(5);			// v3.0 > 16 colours
	wrb(1);			// encoding always 1
	wrb(8);			// 8 bits per pixel, 256 colours
	wrwi(0);			// xmin 0
	wrwi(0);			// ymin 0
	wrwi(width-1);		// xmax
	wrwi(height-1);		// ymax
	wrwi(width);		// hres
	wrwi(height);		// vres

	for (x = 0; x < 48; ++x)
		wrb(0);			// colormap

	wrb(0);			// reserved always 0
	wrb(1);			// numcolourplanes (1 for 8bit)
	wrwi(width);		// bytes per line	
	wrwi(1);			// palette info (1 = color)
	wrwi(width);
	wrwi(height);

	for (x = 0; x < 54; ++x)
		wrb(0);				// pad

	for (x = 0; x < height; ++x)
		pcx_writeline(bmap + (x * width), width);

	pcx_writepalette();
}

static void pcx_writeline(byte *p, word len)
{
	word i;
	byte byt, samect, repcode;

	i = 0;

	do {
		byt = p[i++];
		samect = 1;
		while ((samect < (unsigned)63) && (i < len) && (byt == p[i])) {
			++samect;
			++i;
		}
		if ((samect > 1) || ((byt & 0xc0) != 0)) {
			repcode = 0xC0 | samect;
			wrb(repcode);
		}
		wrb(byt);
	} while (i < len);
}

byte pal[] = {	0x00,0x00,0x00, 0x00,0x00,0x2b, 0x00,0x2b,0x00, 0x00,0x2b,0x2b,
				0x2b,0x00,0x00, 0x2d,0x00,0x2d, 0x2b,0x1b,0x00, 0x2b,0x2b,0x2b,
				0x1b,0x1b,0x1b, 0x1b,0x1b,0x3f, 0x00,0x3f,0x00, 0x00,0x3f,0x3f,
				0x3f,0x1f,0x1f, 0x3f,0x00,0x3f, 0x3f,0x3f,0x00, 0x3f,0x3f,0x3f };

static void pcx_writepalette(void)
{
	register word x;

	wrb(12);
	for (x = 0; x < (16*3); ++x)
		wrb(pal[x] << 2);
	for (x = (16*3); x < 768; ++x)
		wrb(0);
}
